/*
 * Copyright (C) 1997-2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include <sys/time.h> /* gettimeofday */
#include "server.cmdline.h"
#include "db.h"
#include "server.h"
#include "afs.h"
#include "send.h"

extern const char *status_item_list[];

static struct timeval announce_tv;
static struct timeval data_send_barrier;
static struct timeval eof_barrier;

extern struct misc_meta_data *mmd;
extern struct dbtool dblist[];
extern struct sender senders[];
extern struct gengetopt_args_info conf;

static FILE *audio_file = NULL;

#if 1
	void mp3_init(void *);
#endif

#ifdef HAVE_OGGVORBIS
	void ogg_init(void *);
#endif

struct audio_format afl[] = { /* audio format list */
#if 1
	{
		.name = "mp3",
		.init = mp3_init,
	},
#endif
#ifdef HAVE_OGGVORBIS
	{
		.name = "ogg",
		.init = ogg_init,
	},
#endif
	{
		.name = NULL,
	}
};


/**
 * afs_playing - check if audio file sender is playing
 *
 * Returns 1 if playing, 0 otherwise.
 * 
 * TODO: Should be a macro
 */
unsigned int afs_playing(void)
{
	return mmd->new_afs_status_flags & AFS_PLAYING;
}
unsigned int afs_next(void)
{
	return mmd->new_afs_status_flags & AFS_NEXT;
}
unsigned int afs_repos(void)
{
	return mmd->new_afs_status_flags & AFS_REPOS;
}
unsigned int afs_paused(void)
{
	return !(mmd->new_afs_status_flags & AFS_NEXT) 
		&& !(mmd->new_afs_status_flags & AFS_PLAYING);
}

/**
 * audio_format_name - get the name of the given audio format
 * @i: audio format number
 *
 * This returns a pointer to statically allocated memory so it
 * must not be freed by the caller.
 */
const char *audio_format_name(int i)
{
	return i >= 0?  afl[i].name : "(none)";
}

int afs_handle_sighup(void)
{
	return 1;
}

void afs_init(void)
{
	int i;
	char *hn = para_hostname(), *home = para_homedir();

	PARA_DEBUG_LOG("supported audio formats: %s\n",
		SUPPORTED_AUDIO_FORMATS);
	for (i = 0; afl[i].name; i++) {
		PARA_NOTICE_LOG("initializing %s handler\n",
			afl[i].name);
		afl[i].init(&afl[i]);
	}
	ms2tv(conf.announce_time_arg, &announce_tv);
	PARA_INFO_LOG("announce timeval: %lu:%lu\n", announce_tv.tv_sec, announce_tv.tv_usec);
	for (i = 0; senders[i].name; i++) {
		PARA_NOTICE_LOG("initializing %s sender\n", senders[i].name);
		senders[i].init(&senders[i]);
	}
	free(hn);
	free(home);
}

static int get_file_info(int i)
{
	/* get_file_info must not close the audio file, even if it is invalid */
	int ret = afl[i].get_file_info(audio_file, mmd->audio_file_info,
		&mmd->chunks_total, &mmd->seconds_total);
	if (ret > 0)
		PARA_INFO_LOG("valid %s file\n", afl[i].name);
	else
		PARA_DEBUG_LOG("%s format not recognized\n", afl[i].name);
	return ret;
}

static int guess_audio_format(const char *name)
{

	int i, len1 = strlen(name), len2;

	for (i = 0; afl[i].name; i++) {
		len2 = strlen(afl[i].name);
		if (len1 < len2)
			continue;
		if (!strncasecmp(name + (len1 - len2), afl[i].name, len2)) {
			PARA_DEBUG_LOG("might be %s\n", afl[i].name);
			return i;
		}
	}
	return -1;
}

static int get_audio_format(int omit)
{
	int i;

	for (i = 0; afl[i].name; i++) {
		if (i == omit || !afl[i].get_file_info)
			continue;
		rewind(audio_file);
		if (get_file_info(i) > 0)
			return i;
		rewind(audio_file);
	}
	return -1;
}

/*
 * upddate shared mem
 */
static int update_mmd(void)
{
	int i;
	struct stat file_status;

	i = guess_audio_format(mmd->filename);
	if (i < 0 || get_file_info(i) < 0)
		i = get_audio_format(i);
	if (i < 0) {
		PARA_WARNING_LOG("unknown audio format for %s\n",
			mmd->filename);
		return -1;
	}
	mmd->audio_format = i;
	mmd->chunks_sent = 0;
	mmd->current_chunk = 0;
	mmd->offset = 0;
	if (fstat(fileno(audio_file), &file_status) == -1) {
		PARA_ERROR_LOG("fstat error (%s)\n", strerror(errno));
		return -1;
	}
	mmd->size = file_status.st_size;
	mmd->mtime = file_status.st_mtime;
	mmd->events++;
	PARA_NOTICE_LOG("next audio file: %s\n", mmd->filename);
	return 1;
}

static void get_song(void)
{
	char **sl = dblist[mmd->dbt_num].get_audio_file_list(10);
	int i;

	if (!sl)
		goto err_out;
	for (i = 0; sl[i]; i++) {
		struct timeval now;
		PARA_INFO_LOG("trying %s\n", sl[i]);
		if (strlen(sl[i]) >= _POSIX_PATH_MAX)
			continue;
		audio_file = fopen(sl[i], "r");
		if (!audio_file)
			continue;
		strcpy(mmd->filename, sl[i]);
		if (update_mmd() < 0) {
			fclose(audio_file);
			audio_file = NULL;
			continue;
		}
		mmd->num_played++;
		dblist[mmd->dbt_num].update_audio_file(sl[i]);
		PARA_DEBUG_LOG("%s", "success\n");
		mmd->new_afs_status_flags &= (~AFS_NEXT);
		gettimeofday(&now, NULL);
		tv_add(&now, &announce_tv, &data_send_barrier);
	
		goto free;
	}
	PARA_ERROR_LOG("%s", "no valid files found\n");
err_out:
	mmd->new_afs_status_flags = AFS_NEXT;
free:
	if (sl) {
		for (i = 0; sl[i]; i++)
			free(sl[i]);
		free(sl);
	}
}

static int chk_barrier(const char *bname, const struct timeval *now,
		const struct timeval *barrier, struct timeval *diff, int log)
{
	long ms;

	if (tv_diff(now, barrier, diff) > 0)
		return 1;
	ms = tv2ms(diff);
	if (log && ms)
		PARA_DEBUG_LOG("%s barrier: %lims left\n", bname, ms);
	return -1;
}

static void afs_next_chunk_time(struct timeval *due)
{
	struct timeval tmp;

	tv_scale(mmd->chunks_sent, &afl[mmd->audio_format].chunk_tv, &tmp);
	tv_add(&tmp, &mmd->stream_start, due);
}

/*
 * < 0: timeout was written to timeout
 * > 0: nothing to do
 */
static int afs_compute_timeout(struct timeval **timeout)
{
	static struct timeval the_timeout;
	struct timeval now, next_chunk;

	*timeout = &the_timeout;
	if (afs_next() && mmd->audio_format >= 0) {
		/* only sleep a bit, nec*/
		the_timeout.tv_sec = 0;
		the_timeout.tv_usec = 100;
		return -1;
	}
	gettimeofday(&now, NULL);
	if (chk_barrier("eof", &now, &eof_barrier, &the_timeout, 1) < 0)
		return -1;
	if (chk_barrier("data send", &now, &data_send_barrier,
			 &the_timeout, 1) < 0)
		return -1;
	if (mmd->audio_format < 0 || !afs_playing() || !audio_file)
		return 1;
	afs_next_chunk_time(&next_chunk);
	if (chk_barrier(afl[mmd->audio_format].name, &now, &next_chunk,
			&the_timeout, 0) < 0)
		return -1;
	/* chunk is due or bof */
	the_timeout.tv_sec = 0;
	the_timeout.tv_usec = 0;
	return -1;
}

static void afs_eof(struct audio_format *af)
{
	struct timeval now;
	int i;
	char *tmp;

	if (!af || !audio_file) {
		for (i = 0; senders[i].name; i++)
			senders[i].shutdown_clients();
		return;
	}
	gettimeofday(&now, NULL);
	tv_add(&af->eof_tv, &now, &eof_barrier);
	af->close_audio_file();
	audio_file = NULL;
	mmd->audio_format = -1;
	af = NULL;
	mmd->chunks_sent = 0;
	mmd->offset = 0;
	mmd->seconds_total = 0;
	tmp  = make_message("%s:\n%s:\n%s:\n", status_item_list[SI_AUDIO_INFO1],
		status_item_list[SI_AUDIO_INFO2], status_item_list[SI_AUDIO_INFO3]);
	strcpy(mmd->audio_file_info, tmp);
	free(tmp);
	tmp  = make_message("%s:\n%s:\n%s:\n", status_item_list[SI_DBINFO1],
		status_item_list[SI_DBINFO2], status_item_list[SI_DBINFO3]);
	strcpy(mmd->dbinfo, tmp);
	free(tmp);
	mmd->filename[0] = '\0';
	mmd->size = 0;
	mmd->events++;
}

/*
 * called from main in any case, just before main calls select
 */
int afs_mainloop(struct timeval **timeout)
{
	struct audio_format *af = NULL;
	int i, ret, format;
	struct timeval now;

	*timeout = NULL;
again:
	format = mmd->audio_format;
	if (format >= 0)
		af = afl + format;
	else
		for (i = 0; senders[i].name; i++)
			senders[i].shutdown_clients();
	if (afs_next() && af) {
		afs_compute_timeout(timeout);
		afs_eof(af);
		return -1;
	}
	if (afs_paused() || afs_repos()) {
		for (i = 0; senders[i].name; i++)
			senders[i].shutdown_clients();
		if (af) {
			if (!afs_paused() || mmd->chunks_sent) {
				gettimeofday(&now, NULL);
				tv_add(&af->eof_tv, &now, &eof_barrier);
			}
			if (afs_repos())
				tv_add(&now, &announce_tv, &data_send_barrier);
			if (mmd->new_afs_status_flags & AFS_NOMORE)
				mmd->new_afs_status_flags = AFS_NEXT;
		}
		mmd->chunks_sent = 0;
	}
	if (af && afs_repos() && mmd->current_chunk != mmd->repos_request)
		af->reposition_stream(mmd->repos_request);
	ret = -1;
	if (afs_repos()) {
		mmd->new_afs_status_flags &= ~(AFS_REPOS);
		mmd->current_chunk = mmd->repos_request;
	}
	ret = afs_compute_timeout(timeout);
	if (ret < 0)
		goto out;
	if (!audio_file && afs_playing() && 
			!(mmd->new_afs_status_flags & AFS_NOMORE)) {
		PARA_DEBUG_LOG("%s", "ready and playing, but no audio file\n");
		get_song();
		goto again;
	}
	ret = -1;
out:
//	PARA_DEBUG_LOG("%s: returning fd %d, tv = %lims\n", __func__,
//		ret, *timeout? tv2ms(*timeout) : 0);

	return ret;
}

int afs_send_chunk(void)
{
	int i, ret;
	struct audio_format *af;
	char *buf;
	size_t len;
	struct timeval now, due;

	if (mmd->audio_format < 0 || !audio_file || !afs_playing())
		return -1;
	af = &afl[mmd->audio_format];
	gettimeofday(&now, NULL);
	afs_next_chunk_time(&due);
	if (tv_diff(&due, &now, NULL) > 0)
		return -1;
	ret = af->read_chunk(mmd->current_chunk, &len, &buf);
	mmd->new_afs_status_flags &= ~(AFS_NEXT | AFS_REPOS);
	if (ret > 0) {
		if (!mmd->chunks_sent) {
			struct timeval tmp;
			gettimeofday(&mmd->stream_start, NULL);
			tv_scale(mmd->current_chunk, &af->chunk_tv, &tmp);
			mmd->offset = tv2ms(&tmp);
			mmd->events++;
		}
		for (i = 0; senders[i].name; i++)
			senders[i].send(af, mmd->current_chunk,
				mmd->chunks_sent, buf, len);
		mmd->new_afs_status_flags |= AFS_PLAYING;
		mmd->chunks_sent++;
		mmd->current_chunk++;
	}
	if (ret < 0)
		mmd->new_afs_status_flags = AFS_NEXT;
	if (!ret)
		mmd->new_afs_status_flags |= AFS_NEXT;
	if (ret <= 0)
		afs_eof(af);
	return ret;
}
