/*
 * Copyright (C) 1997-2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */


#include <malloc.h> /* mallinfo */
#include <sys/time.h> /* gettimeofday */
#include "crypt.h"
#include "server.cmdline.h"
#include "db.h"
#include "server.h"
#include "afs.h"
#include "send.h"
#include "rc4.h"
#include <openssl/rc4.h>
#include "gcc-compat.h"

void (*crypt_function_recv)(unsigned long len, const unsigned char *indata, unsigned char *outdata) = NULL;
void (*crypt_function_send)(unsigned long len, const unsigned char *indata, unsigned char *outdata) = NULL;
static RC4_KEY rc4_recv_key;
static RC4_KEY rc4_send_key;
static unsigned char rc4_buf[2 * RC4_KEY_LEN];

extern const char *status_item_list[NUM_STAT_ITEMS];
extern struct misc_meta_data *mmd;
extern struct gengetopt_args_info conf;
extern struct dbtool dblist[];
extern struct audio_format afl[];
extern struct sender senders[];
extern char *user_list;
struct sockaddr_in *in_addr;

static int com_si(int, int, char **);
static int com_version(int, int, char **);
static int com_sb(int, int, char **);
static int com_sc(int, int, char **);
static int com_stat(int, int, char **);
static int com_help(int, int, char **);
static int com_hup(int, int, char **);
static int com_term(int, int, char **);
static int com_play(int, int, char **);
static int com_stop(int, int, char **);
static int com_pause(int, int, char **);
static int com_next(int, int, char **);
static int com_nomore(int, int, char **);
static int com_cdt(int, int, char **);
static int com_ff(int, int, char **);
static int com_jmp(int, int, char **);
static int com_sender(int, int, char **);


enum {
	E_GENERAL,	E_SYNTAX,	E_SEND,		E_RECV,
	E_NODBTOOL,	E_NOAUDIOFILE,	E_FORK,		E_PIPE,
	E_AUTH,		E_CMD_INVALID,	E_PERM,		E_SPLIT,
	E_RAND,		E_USERLIST,	E_NOUSER,	E_LOCK,
	E_NOT_SUPP,	E_RC4,		E_ENCRYPT
};
static const char *internal_err_list[] = {
	"general error",
	"syntax error",
	"send error",
	"receive error",

	"no such database tool",
	"no audio file",
	"can not fork",
	"pipe error",

	"authentication error",
	"invalid command",
	"permission denied",
	"can not split command line args",

	"urandom error",
	"problems reading user list",
	"you don't exist. Go away.",
	"lock error",

	"command not supported",
	"rc4 error",
	"crypt error"
};
/* commands that are handled by the server itself */
static struct command cmd_struct[] = {
{
.name = "cdt",
.handler = com_cdt,
.perms = DB_READ | DB_WRITE,
.description = "change database tool",
.usage = "cdt [name_of_new_dbtool]",
.help =
"Deactivate current dbtool and activate name_of_new_dbtool. If no\n"
"argument was given, print the current database tool.\n"
},

{
.name = "ff",
.handler = com_ff,
.perms = AFS_READ | AFS_WRITE,
.description = "jmp amount of time forwards or backwards "
	"in current audio file",
.usage = "ff n[-]",
.help =

"\tSet the 'R' (reposition request) bit of the afs status flags\n"
"\tand enqueue a request to jump n seconds forwards or backwards\n"
"\tin the current audio file.\n"
"\n"
"EXAMPLE\n"
"\n"
"\t\tff 30-\n"
"\n"
"\tjumps 30 seconds backwards.\n"

},

{
.name = "help",
.handler = com_help,
.perms = 0,
.description = "print help text",
.usage = "help [command]",
.help =

"Without any arguments, help prints a list of availible commands. When\n"
"issued with a command name as first argument, print out a description\n"
"for that command.\n"

},

{
.name = "hup",
.handler = com_hup,
.perms = AFS_WRITE,
.description = "force reload of config file and log file",
.usage = "hup",
.help =

"After rereading the config file, a signal is sent to all children\n"
"which forces them to close/reopen the log file.\n"

},

{
.name = "jmp",
.handler = com_jmp,
.perms = AFS_READ | AFS_WRITE,
.description = "jmp to given position in current audio file",
.usage = "jmp [n]",
.help =

"\tSet the 'R' (reposition request) bit of the afs status flags\n"
"\tand enqueue a request to jump to n% of the current audio file,\n"
"\twhere 0 <= n <= 100.\n"

},

{
.name = "next",
.handler = com_next,
.perms = AFS_READ | AFS_WRITE,
.description = "skip rest of current audio file",
.usage = "next",
.help =

"\tSet the 'N' (next audio file) bit of the afs status flags. When\n"
"\tplaying, change audio file immediately. Equivalent to stop\n"
"\tif paused, NOP if stopped.\n"


},

{
.name = "nomore",
.handler = com_nomore,
.perms = AFS_READ | AFS_WRITE,
.description = "stop playing after current audio file",
.usage = "nomore",
.help =

"Set the 'O' (no more) bit of the afs status flags. This instructs\n"
"para_server to clear the 'P' (playing) bit as soon as it encounters\n"
"the 'N' (next audio file) bit being set.\n"
"\n"
"Use this command instead of stop if you don't like\n"
"sudden endings.\n"

},

{
.name ="pause",
.handler = com_pause,
.perms = AFS_READ | AFS_WRITE,
.description = "pause current audio file",
.usage = "pause",
.help =

"\tClear the 'P' (playing) bit of the afs status flags.\n"

},

{
.name = "play",
.handler = com_play,
.perms = AFS_READ | AFS_WRITE,
.description = "start playing or resume playing when paused",
.usage = "play",
.help =

"\tSet the 'P' (playing) bit of the afs status flags. This\n"
"\tresults in starting/continuing to stream.\n"

},

{
.name = "sb",
.handler = com_sb,
.perms = AFS_READ,
.description = "print status bar for current audio file",
.usage = "sb [n]",
.help =

"Without any arguments, sb continuously prints a status bar of the form\n"
"\n"
"	12:34 [56:12] (56%) filename\n"
"\n"
"indicating playing time, remaining time, percentage and the name of\n"
"the file beeing streamed. Use the optional number n to let stat exit\n"
"after having displayed the status bar n times.\n"

},
{
.name = "sc",
.handler = com_sc,
.perms = AFS_READ,
.description = "print name of audio file whenever it changes",
.usage = "sc [n]",
.help =

"\tsc prints exactly one line (the filename of the audio file\n"
"\tbeing played) whenever the audio file changes. Stops after\n"
"\tn iterations, or never if n is not specified.\n"

},
{
.name = "sender",
.handler = com_sender,
.perms = AFS_READ | AFS_WRITE,
.description = "control paraslash internal senders",
.usage = "sender [s cmd [arguments]]",
.help =

"send command cmd to sender s. cmd may be one of the following:\n"
"help, on, off, add, delete, allow, or deny. Note that not all senders\n"
"support each command. Try e.g. 'para_client sender http help' for\n"
"more information about the http sender. If no argument is given,\n"
"print out a list of all senders that are compiled in.\n"

},
{
.name = "si",
.handler = com_si,
.perms = 0,
.description = "print server info",
.usage = "si",
.help =
"Print server uptime and other information.\n"
},

{
.name = "stat",
.handler = com_stat,
.perms = AFS_READ,
.description = "print status info for current audio file",
.usage = "stat [n]",
.help =

"\tWithout any arguments, stat continuously prints status messages\n"
"\tof the audio file being streamed. Use the optional number n\n"
"\tto let stat exit after having displayed status n times.\n"

},

{
.name = "stop",
.handler = com_stop,
.perms = AFS_READ | AFS_WRITE,
.description = "stop playing",
.usage = "stop",
.help =

"\tClear the 'P' (play) bit and set the 'N' bit of the afs status\n"
"\tflags.\n"

},
{
.name = "term",
.handler = com_term,
.perms = AFS_READ | AFS_WRITE,
.description = "terminate para_server",
.usage = "term",
.help =

"Shuts down the server. Instead of this command, you can also send\n"
"SIGINT or SIGTERM. It should never be necessary to send SIGKILL.\n"

},
{
.name = "version",
.handler = com_version,
.perms = 0,
.description = "print server's version",
.usage = "version",
.help =
"Show version and other info\n"
},
/* this indicates the end of the list. Do not touch. */
{
.name = NULL,
}
};

static void dummy(__unused int s)
{}

static void mmd_dup(struct misc_meta_data *new_mmd)
{
	mmd_lock();
	*new_mmd = *mmd;
	mmd_unlock();
}

/*
 * compute human readable string containing
 * afs_status for given integer value
 */
static char *afs_status_tohuman(unsigned int flags)
{
	if (flags & AFS_PLAYING)
		return para_strdup("playing");
	else if (flags & AFS_NEXT)
		return para_strdup("stopped");
	else
		return para_strdup("paused");
}


/*
 * return human readable permission string. Never returns NULL.
 */
char *cmd_perms_itohuman(unsigned int perms)
{
	char *msg = para_malloc(7 * sizeof(char));

	msg[0] = perms & DB_READ? 'd' : '-';
	msg[1] = perms & DB_WRITE? 'D' : '-';
	msg[2] = perms & AFS_READ? 'a' : '-';
	msg[3] = perms & AFS_WRITE? 'A' : '-';
	msg[4] = '\0';
	return msg;
}

/*
 * Never returns NULL.
 */
static char *afs_get_status_flags(unsigned int flags)
{
	char *msg = para_malloc(5 * sizeof(char));

	msg[0] = (flags & AFS_PLAYING)? 'P' : '_';
	msg[1] = (flags & AFS_NOMORE)? 'O' : '_';
	msg[2] = (flags & AFS_NEXT)? 'N' : '_';
	msg[3] = (flags & AFS_REPOS)? 'R' : '_';
	msg[4] = '\0';
	return msg;
}

/*
 * compute status bar string. Never returns NULL
 */
char *get_sb_string(struct misc_meta_data *nmmd)
{
	char *base, *ret;
	long long unsigned secs = 0, rsecs = 0, percent = 0;

	base = para_basename(nmmd->filename);
	if (!base)
		return para_strdup("");
	if (!base[0])
		return base;
	if (nmmd->chunks_total) {
		secs = (long long) nmmd->seconds_total * nmmd->chunks_sent
			/ nmmd->chunks_total;
		rsecs = (long long) nmmd->seconds_total *
			(nmmd->chunks_total - nmmd->chunks_sent)
			/ nmmd->chunks_total;
		percent = 100 * ((nmmd->chunks_sent + 5) / 10)
			/ ((nmmd->chunks_total + 5) / 10);
	}
	ret = make_message("%llu:%02llu [%llu:%02llu] (%llu%%) %s",
		secs / 60, secs % 60,
		rsecs / 60, rsecs % 60,
		percent,
		base
	);
	free(base);
	return ret;
}

static char *get_status(struct misc_meta_data *nmmd)
{
	char *bar, *ret, mtime[30] = "";
	char *status, *flags; /* afs status info */
	char *ut = uptime_str(server_uptime(GET));
	long offset = (nmmd->offset + 500) / 1000;
	struct timeval now;
	struct tm mtime_tm;

	if (nmmd->audio_format >= 0) {
		localtime_r(&nmmd->mtime, &mtime_tm);
		strftime(mtime, 29, "%a %b %d %Y", &mtime_tm);
	}
	/* report real status */
	status = afs_status_tohuman(nmmd->afs_status_flags);
	flags = afs_get_status_flags(nmmd->afs_status_flags);
	bar = para_basename(nmmd->filename);
	gettimeofday(&now, NULL);
	ret = make_message(
		"%s:%lu\n"	"%s:%s\n"		"%s:%i\n"	"%s:%u\n"
		"%s:%s\n"	"%s:%s\n"	"%s:%s\n"	"%s:%s\n"
		"%s:%li\n" 	"%s:%s\n"	"%s"		"%s"
		"%s:%s\n"	"%s:%lu.%lu\n"	"%s:%lu.%lu\n",
		status_item_list[SI_FILE_SIZE], nmmd->size / 1024,
		status_item_list[SI_MTIME], mtime,
		status_item_list[SI_LENGTH], nmmd->seconds_total,
		status_item_list[SI_NUM_PLAYED], nmmd->num_played,

		status_item_list[SI_STATUS_BAR], bar ? bar : "(none)",
		status_item_list[SI_STATUS], status,
		status_item_list[SI_STATUS_FLAGS], flags,
		status_item_list[SI_DBTOOL], dblist[nmmd->dbt_num].name,

		status_item_list[SI_OFFSET], offset,
		status_item_list[SI_FORMAT], audio_format_name(nmmd->audio_format),
		nmmd->dbinfo,
		nmmd->audio_file_info,

		status_item_list[SI_UPTIME], ut,
		status_item_list[SI_STREAM_START], nmmd->stream_start.tv_sec,
			nmmd->stream_start.tv_usec,
		status_item_list[SI_CURRENT_TIME], now.tv_sec, now.tv_usec

	);
	free(bar);
	free(flags);
	free(status);
	free(ut);
	return ret;
}

static int check_sender_args(int argc, char **argv, struct sender_command_data *scd)
{
	int i;
	/* this has to match sender.h */
	const char *subcmds[] = {"add", "delete", "allow", "deny", "on", "off", NULL};

	scd->sender_num = -1;
	if (argc < 0)
		return -E_SYNTAX;
	for (i = 0; senders[i].name; i++)
		if (!strcmp(senders[i].name, argv[0]))
			break;
//	PARA_DEBUG_LOG("%d:%s\n", argc, argv[0]);
	if (!senders[i].name)
		return -E_SYNTAX;
	scd->sender_num = i;
	for (i = 0; subcmds[i]; i++)
		if (!strcmp(subcmds[i], argv[1]))
			break;
	if (!subcmds[i])
		return -E_SYNTAX;
	scd->cmd_num = i;
	scd->self = *in_addr;
	mmd_lock();
	if (!senders[scd->sender_num].client_cmds[scd->cmd_num]) {
		mmd_unlock();
		return -E_NOT_SUPP;
	}
	mmd_unlock();
	switch (scd->cmd_num) {
	case SENDER_ON:
	case SENDER_OFF:
		if (argc != 1)
			return -E_SYNTAX;
		break;
	case SENDER_DENY:
	case SENDER_ALLOW:
		if (argc != 2 && argc != 3)
			return -E_SYNTAX;
		if (!inet_aton(argv[2], &scd->addr))
			return -E_SYNTAX;
		scd->netmask = 32;
		if (argc == 3) {
			scd->netmask = atoi(argv[3]);
			if (scd->netmask < 0 || scd->netmask > 32)
				return -E_SYNTAX;
		}
		break;
	case SENDER_ADD:
	case SENDER_DELETE:
		if (argc != 2 && argc != 3)
			return -E_SYNTAX;
		if (!inet_aton(argv[2], &scd->addr))
			return -E_SYNTAX;
		scd->port = -1;
		if (argc == 3) {
			scd->port = atoi(argv[3]);
			if (scd->port < 0 || scd->port > 65535)
				return -E_SYNTAX;
		}
		break;
	default:
		return -E_SYNTAX;
	}
	return 1;
}

static int com_sender(int fd, int argc, char **argv)
{
	int i, ret;
	struct sender_command_data scd;

	if (!argc) {
		char *msg = NULL;
		for (i = 0; senders[i].name; i++) {
			char *tmp = make_message("%s%s\n",
				msg? msg : "", senders[i].name);
			free(msg);
			msg = tmp;
		}
		ret = send_buffer(fd, msg) < 0? -E_SEND : 1;
		free(msg);
		return ret;
	}
	ret = check_sender_args(argc - 1, argv + 1, &scd);
	if (ret < 0) {
		char *msg;
		if (scd.sender_num < 0)
			return ret;
		msg = senders[scd.sender_num].help();
		send_buffer(fd, msg);
		free(msg);
		return 1;
	}
	for (i = 0; i < 10; i++) {
		mmd_lock();
		if (mmd->sender_cmd_data.cmd_num >= 0) {
			mmd_unlock();
			usleep(100 * 1000);
			continue;
		}
		mmd->sender_cmd_data = scd;
		mmd_unlock();
		break;
	}
	return (i < 10)? 1 : -E_LOCK;
}

/* server info */
static int com_si(int fd, int argc, __unused char **argv)
{
	int i, ret;
	char *ut;
	char *dbtools = NULL, *sender_info = NULL, *sender_list = NULL;
	struct mallinfo mi = mallinfo();

	if (argc)
		return -E_SYNTAX;
	mmd_lock();
	for (i = 0; dblist[i].name; i++) {
		dbtools = para_strcat(dbtools, dblist[i].name);
		dbtools = para_strcat(dbtools, " ");
	}
	for (i = 0; senders[i].name; i++) {
		char *info = senders[i].info();
		sender_info = para_strcat(sender_info, info);
		free(info);
		sender_list = para_strcat(sender_list, senders[i].name);
		sender_list = para_strcat(sender_list, " ");
	}
	ut = uptime_str(server_uptime(GET));
	ret = send_va_buffer(fd, "up: %s\nplayed: %u\n"
		"pid: %d\n"
		"mallinfo: %d\n"
		"connections (active/accepted/total): %u/%u/%u\n"
		"current loglevel: %i\n"
		"supported database tools: %s\n"
		"supported audio formats: %s\n"
		"supported senders: %s\n"
		"%s",
		ut, mmd->num_played,
		getppid(),
		mi.arena / 1024,
		mmd->active_connections,
		mmd->num_commands,
		mmd->num_connects,
		conf.loglevel_arg,
		dbtools,
		SUPPORTED_AUDIO_FORMATS,
		sender_list,
		sender_info
	) < 0? -E_SEND : 1;
	mmd_unlock();
	free(ut);
	free(dbtools);
	free(sender_list);
	free(sender_info);
	return ret;
}

/* version */
static int com_version(int socket_fd, int argc, __unused char **argv)
{
	if (argc)
		return -E_SYNTAX;
	return send_buffer(socket_fd, "para_server-" VERSION ", \"" CODENAME "\"\n"
			COPYRIGHT "\n"
			"built: " BUILD_DATE "\n"
			SYSTEM ", " CC_VERSION "\n"
		) < 0? -E_SEND : 1;
}

/* sc */
static int com_sc(int socket_fd, int argc, char **argv)
{
	char *name = NULL;
	int ret, old = 0, count = -1; /* print af change forever */

	if (argc)
		count = atoi(argv[1]);
repeat:
	mmd_lock();
	if (old != mmd->num_played) {
		old = mmd->num_played;
		name = para_strdup(mmd->filename);
	}
	mmd_unlock();
	if (name) {
		ret = send_va_buffer(socket_fd, "%s\n", name);
		free(name);
		name = NULL;
		if (ret < 0)
			return -E_SEND;
		if (argc && !--count)
			return 1;
	}
	usleep(500000);
	goto repeat;
}

/* sb */
static int com_sb(int socket_fd, int argc, char **argv)
{
	char *sb;
	int ret, nr = -1;	/* status bar will be printed that many
				 * times. Negative value means: print
				 * forever
				 */
	if (argc)
		nr = atoi(argv[1]);
	while (nr) {
		mmd_lock();
		sb = get_sb_string(mmd);
		mmd_unlock();
		ret = send_va_buffer(socket_fd, "%s\n", sb);
		free(sb);
		if (ret < 0)
			return -E_SEND;
		if (nr == 1)
			return 1;
		usleep(500000);
		if (nr > 0)
			nr--;
	}
	return 1;
}

/* stat */
static int com_stat(int socket_fd, int argc, char **argv)
{
//	char *old_stat = NULL, *old_dbinfo = NULL;
	int ret, num = 0;/* status will be printed that many
			  * times. num <= 0 means: print forever
			  */
	struct misc_meta_data tmp, *nmmd = &tmp;
	char *s;

	signal(SIGUSR1, dummy);

	if (argc)
		num = atoi(argv[1]);
	for (;;) {

		mmd_dup(nmmd);
		s = get_status(nmmd);
		ret = send_buffer(socket_fd, s) < 0?
			-E_SEND : 1;
		free(s);
		if (ret < 0)
			goto out;
		ret = 1;
		if (num == 1)
			goto out;
		usleep(500000 * 100);
	}
out:
	return ret;
}

static int send_description(int fd, struct command *cmd, const char *handler, int num)
{
	int ret, i;

	for (i = 1; cmd->name && (!num || i <= num); cmd++, i++) {
		char *perms = cmd_perms_itohuman(cmd->perms);
		ret = send_va_buffer(fd, "%s\t%s\t%s\t%s\n", cmd->name,
			handler,
			perms,
			cmd->description) < 0? -E_SEND : 1;
		free(perms);
		if (ret < 0)
			return ret;
	}
	return 1;
}

/* always returns string that must be freed by the caller in handeler */
static struct command *get_cmd_ptr(char *name, char **handler, const char ***err_list)
{
	struct command *cmd = cmd_struct;

	if (err_list)
		*err_list = internal_err_list;
	for (cmd = cmd_struct; cmd->name; cmd++)
		if (!strcmp(cmd->name, name)) {
			if (handler)
				*handler = para_strdup("para_server"); /* server commands */
			return cmd;
		}
	/* not found, look for dbtool commands */
	mmd_lock();
	if (handler)
		*handler = make_message("the %s database tool", dblist[mmd->dbt_num].name);
	if (err_list)
		*err_list = dblist[mmd->dbt_num].err_list;
	cmd = dblist[mmd->dbt_num].cmd_list;
	mmd_unlock();
	for (; cmd->name; cmd++)
		if (!strcmp(cmd->name, name))
			return cmd;
	return NULL;
}

/* help */
static int com_help(int fd, int argc, char **argv)
{
	struct command *cmd;
	char *perms, *handler;
	int ret;

	if (!argc) {
		/* no argument given, print list of commands */
		if ((ret = send_description(fd, cmd_struct, "server", 0)) < 0)
			return ret;
		mmd_lock();
		handler = para_strdup(dblist[mmd->dbt_num].name);
		cmd = dblist[mmd->dbt_num].cmd_list;
		mmd_unlock();
		ret = send_description(fd, cmd, handler, 0);
		free(handler);
		return ret;
	}
	/* argument given for help */
	cmd = get_cmd_ptr(argv[1], &handler, NULL);
	if (!cmd) {
		free(handler);
		return -E_CMD_INVALID;
	}
	perms = cmd_perms_itohuman(cmd->perms);
	ret = send_va_buffer(fd,
		"NAME\n\t%s - %s\n"
		"SYNOPSIS\n\t para_client %s\n"
		"DESCRIPTION\n%s\n"
		"HANDLER\n"
		"This command is handled by %s.\n\n"
		"PERMISSIONS\n"
		"Needed privileges for %s: %s\n",
		argv[1],
		cmd->description,
		cmd->usage,
		cmd->help,
		handler,
		argv[1],
		perms
	) < 0? -E_SEND : 0;
	free(perms);
	free(handler);
	return ret;
}

/* hup */
static int com_hup(__unused int socket_fd, int argc, __unused char **argv)
{
	if (argc)
		return -E_SYNTAX;
	kill(getppid(), SIGHUP);
	return 1;
}

/* term */
static int com_term(__unused int socket_fd, int argc, __unused char **argv)
{
	if (argc)
		return -E_SYNTAX;
	kill(getppid(), SIGTERM);
	return 1;
}

static int com_play(__unused int socket_fd, int argc, __unused char **argv)
{
	if (argc)
		return -E_SYNTAX;
	mmd_lock();
	mmd->new_afs_status_flags |= AFS_PLAYING;
	mmd->new_afs_status_flags &= ~AFS_NOMORE;
	mmd_unlock();
	return 1;

}

/* stop */
static int com_stop(__unused int socket_fd, int argc, __unused char **argv)
{
	if (argc)
		return -E_SYNTAX;
	mmd_lock();
	mmd->new_afs_status_flags &= ~AFS_PLAYING;
	mmd->new_afs_status_flags &= ~AFS_REPOS;
	mmd->new_afs_status_flags |= AFS_NEXT;
	mmd_unlock();
	return 1;
}

/* pause */
static int com_pause(__unused int socket_fd, int argc, __unused char **argv)
{
	if (argc)
		return -E_SYNTAX;
	mmd_lock();
	if (!afs_paused())
		mmd->events++;
	mmd->new_afs_status_flags &= ~AFS_PLAYING;
	mmd->new_afs_status_flags &= ~AFS_NEXT;
	mmd_unlock();
	return 1;
}

static int com_cdt(int fd, int argc, char **argv)
{
	int i, ret;

	if (!argc) {
		char *dbtool;
		mmd_lock();
		dbtool = para_strdup(dblist[mmd->dbt_num].name);
		mmd_unlock();
		ret = send_va_buffer(fd, "%s\n", dbtool) < 0?
			-E_SEND : 1;
		free(dbtool);
		return ret;
	}
	for (i = 0; dblist[i].name; i++) {
		if (strcmp(dblist[i].name, argv[1]))
			continue;
		mmd_lock();
		mmd->dbt_change = i;
		mmd->events++;
		mmd_unlock();
		return 1;
	}
	return -E_NODBTOOL;
}

/* next */
static int com_next(__unused int socket_fd, int argc, __unused char **argv)
{
	if (argc)
		return -E_SYNTAX;
	mmd_lock();
	mmd->new_afs_status_flags |= AFS_NEXT;
	mmd_unlock();
	return 1;
}

/* nomore */
static int com_nomore(__unused int socket_fd, int argc, __unused char **argv)
{
	if (argc)
		return -E_SYNTAX;
	mmd_lock();
	mmd->new_afs_status_flags |= AFS_NOMORE; /* FIXME: NOP if !playing? */
	mmd_unlock();
	return 1;
}

/* ff */
static int com_ff(__unused int socket_fd, int argc, char **argv)
{
	long promille;
	int ret, backwards = 0;
	unsigned i;
	char c;

	if (!argc)
		return -E_SYNTAX;
	if (!(ret = sscanf(argv[1], "%u%c", &i, &c)))
		return -E_SYNTAX;
	if (ret > 1 && c == '-')
		backwards = 1; /* jmp backwards */
	mmd_lock();
	ret = -E_NOAUDIOFILE;
	if (!mmd->chunks_total || !mmd->seconds_total)
		goto out;
	promille = (1000 * mmd->current_chunk) / mmd->chunks_total;
	if (backwards)
		promille -= 1000 * i / mmd->seconds_total;
	else
		promille += 1000 * i / mmd->seconds_total;
	if (promille < 0)
		promille = 0;
	if (promille >  1000) {
		mmd->new_afs_status_flags |= AFS_NEXT;
		goto out;
	}
	mmd->repos_request = (mmd->chunks_total * promille) / 1000;
	mmd->new_afs_status_flags |= AFS_REPOS;
	mmd->new_afs_status_flags &= ~AFS_NEXT;
	mmd->events++;
	ret = 1;
out:
	mmd_unlock();
	return ret;
}

/* jmp */
static int com_jmp(__unused int socket_fd, int argc, char **argv)
{
	long unsigned int i;
	int ret;

	if (!argc)
		return -E_SYNTAX;
	if (sscanf(argv[1], "%lu", &i) <= 0)
		return -E_SYNTAX;
	mmd_lock();
	ret = -E_NOAUDIOFILE;
	if (!mmd->chunks_total)
		goto out;
	if (i > 100)
		i = 100;
	PARA_INFO_LOG("jumping to %lu%%\n", i);
	mmd->repos_request = (mmd->chunks_total * i + 50)/ 100;
	PARA_INFO_LOG("sent: %lu,  offset before jmp: %lu\n",
		mmd->chunks_sent, mmd->offset);
	mmd->new_afs_status_flags |= AFS_REPOS;
	mmd->new_afs_status_flags &= ~AFS_NEXT;
	ret = 1;
	mmd->events++;
out:
	mmd_unlock();
	return ret;
}

/*
 * check if perms are sufficient to exec a command having perms cmd_perms.
 * Returns 0 if perms are sufficient, -E_PERM otherwise.
 */
static int check_perms(unsigned int perms, struct command *cmd_ptr)
{
	PARA_DEBUG_LOG("%s", "checking permissions\n");
	return (cmd_ptr->perms & perms) < cmd_ptr->perms ? -E_PERM : 0;
}

/*
 * Parse first string from *cmd and lookup in table of valid commands.
 * On error, NULL is returned.
 */
static struct command *parse_cmd(const char *cmdstr, const char ***err_list)
{
	char buf[255];
	struct command *cmd = NULL;
	int n = 0;

	sscanf(cmdstr, "%200s%n", buf, &n);
	if (!n)
	        return NULL;
	buf[n] = '\0';
	cmd = get_cmd_ptr(buf, NULL, err_list);
	if (cmd)
		PARA_DEBUG_LOG("found command %s\n", cmd->name);
	else
		PARA_ERROR_LOG("unknown command: %s\n", buf);
	return cmd;
}

long int para_rand(long unsigned max)
{
	return (long int) ((max + 0.0) * (random() / (RAND_MAX + 1.0)));
}


/* Open user_list file, returns pointer to opened file on success,
 * NULL on errors
 */
static FILE *open_user_list(char *file)
{
	FILE *cf;

	PARA_DEBUG_LOG("opening user list %s\n", file);
	if (!(cf = fopen(file, "r"))) {
		PARA_INFO_LOG("failed to open user list %s (%s)\n",
			file, strerror(errno));
	        return NULL;
	}
	return cf;
}

/*
 * lookup user in user_list file. Fills in a user struct containing
 * filename of the user's public key as well as the permissions of that user.
 * Returns 1 on success, 0 if user does not exist and < 0 on errors.
 */
static int get_user(struct user *user) {
	FILE *file_ptr;
	char *char_ptr;
	char line[MAXLINE];
	/* keyword, user, key, perms */
	char w[MAXLINE], n[MAXLINE], k[MAXLINE], p[MAXLINE], tmp[4][MAXLINE];
	int num;

	file_ptr = open_user_list(user_list);
	if (!file_ptr)
		return -E_USERLIST;
	while (fgets(line, MAXLINE, file_ptr)) {
//		PARA_DEBUG_LOG("%s: Read line (%i bytes) "
//			"from config file\n", __func__, strlen(line));
		if (sscanf(line,"%200s %200s %200s %200s", w, n, k, p) < 3)
			continue;
		if (!strcmp(w, "user") && !strcmp(user->name, n)) {
			PARA_DEBUG_LOG("found entry for %s\n", n);
			strcpy(user->name, n);
			strcpy(user->pubkey_file, k);
			user->perms = 0;
			char_ptr = p;
			num = sscanf(char_ptr, "%200[A-Z_],%200[A-Z_],%200[A-Z_],%200[A-Z_]",
				tmp[0], tmp[1], tmp[2], tmp[3]);
			PARA_DEBUG_LOG("found %i perm entries\n",
				num);
			user->perms = 0;
			while (num > 0) {
				num--;
				//PARA_DEBUG_LOG("%s: tmp[%i]=%s\n", __func__,
				//	num, tmp[num]);
				if (!strcmp(tmp[num], "AFS_READ"))
					user->perms =
						user->perms | AFS_READ;
				else if (!strcmp(tmp[num], "AFS_WRITE"))
					user->perms =
						user->perms | AFS_WRITE;
				else if (!strcmp(tmp[num], "DB_READ"))
					user->perms = user->perms | DB_READ;
				else if (!strcmp(tmp[num], "DB_WRITE"))
					user->perms = user->perms | DB_WRITE;
				else /* unknown permission */
					PARA_WARNING_LOG("unknown permission:"
						"%s\n", tmp[num]);
			}
			fclose(file_ptr);
			return 1;
		}
	}
	fclose(file_ptr);
	return 0;
}

static void init_rc4_keys(void)
{
	int i;

	for (i = 0; i < 2 * RC4_KEY_LEN; i++)
		rc4_buf[i] = para_rand(256);
	PARA_DEBUG_LOG("rc4 keys initialized (%u:%u)\n",
		(unsigned char) rc4_buf[0],
		(unsigned char) rc4_buf[RC4_KEY_LEN]);
        RC4_set_key(&rc4_recv_key, RC4_KEY_LEN, rc4_buf);
        RC4_set_key(&rc4_send_key, RC4_KEY_LEN, rc4_buf + RC4_KEY_LEN);
}

static void rc4_recv(unsigned long len, const unsigned char *indata, unsigned char *outdata)
{
        RC4(&rc4_recv_key, len, indata, outdata);
}
static void rc4_send(unsigned long len, const unsigned char *indata, unsigned char *outdata)
{
        RC4(&rc4_send_key, len, indata, outdata);
}




int handle_connect(int fd, struct sockaddr_in *addr)
{
	int numbytes, ret, argc, use_rc4 = 0;
	char buf[STRINGSIZE];
	unsigned char crypt_buf[MAXLINE];
	struct user u;
	struct command *cmd = NULL;
	long unsigned challenge_nr, chall_response;
	char **argv;
	const char **err_list = internal_err_list;
	char *p, *command = NULL;

	signal(SIGCHLD, SIG_IGN);
	signal(SIGINT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);
	signal(SIGHUP, SIG_DFL);
	signal(SIGUSR1, SIG_IGN);

	in_addr = addr;
	ret = -E_RAND;
	challenge_nr = random();
	/* send Welcome message */
	ret = -E_SEND;
	if (send_va_buffer(fd, "This is para_server, version " VERSION  ".\n" ) < 0)
		goto err_out;
	/* recv auth request line */
	ret = -E_RECV;
	if ((numbytes = recv_buffer(fd, buf, sizeof(buf))) <= 6)
		goto err_out;
	ret = -E_AUTH;
	if (strncmp(buf, "auth ", 5))
		goto err_out;

	if (numbytes < 9 || strncmp(buf, "auth rc4 ", 9))
		strcpy(u.name, buf + 5); /* client version < 0.2.6 */
	else {
		strcpy(u.name, buf + 9); /* client version >= 0.2.6 */
		use_rc4 = 1;
	}
//	strcpy(u.name, buf + 5); /* ok, but ugly */
	PARA_DEBUG_LOG("received %s request for user %s\n",
		use_rc4? "rc4" : "auth", u.name);
	ret = -E_USERLIST;
	/* lookup user in list file */
	if ((ret = get_user(&u)) < 0)
		goto err_out;
	if (!ret) { /* user not found */
		PARA_WARNING_LOG("auth request for unknown user %s\n", u.name);
		ret = -E_NOUSER;
		goto err_out;
	}
	if ((ret = para_encrypt(u.pubkey_file, challenge_nr, crypt_buf)) <= 0) {
		ret = -E_ENCRYPT;
		goto err_out;
	}
	numbytes = ret;
	PARA_DEBUG_LOG("sending %d byte challenge\n", numbytes);
	/* We can't use send_buffer here since buf may contain null bytes */
	ret = -E_SEND;
	if (send_bin_buffer(fd,(char *) crypt_buf, numbytes) < 0)
		goto err_out;
	/* recv decrypted number */
	ret = -E_RECV;
	if ((numbytes = recv_buffer(fd, buf, sizeof(buf))) <= 0)
		goto err_out;
	ret = -E_AUTH;
	if (sscanf(buf, CHALLENGE_RESPONSE_MSG "%lu", &chall_response) < 1
			|| chall_response != challenge_nr) {
		PARA_WARNING_LOG("%s", "bad auth, aborting\n");
		goto err_out;
	}
	/* auth successful. Send 'Proceed' message */
	PARA_INFO_LOG("good auth for %s (%lu)\n", u.name, challenge_nr);
	sprintf(buf, "%s", PROCEED_MSG);
	if (use_rc4) {
		ret = -E_RC4;
		init_rc4_keys();
		ret = para_encrypt_buffer(u.pubkey_file, rc4_buf, 2 * RC4_KEY_LEN,
			(unsigned char *)buf + PROCEED_MSG_LEN + 1);
		if (ret <= 0) {
			ret = -E_ENCRYPT;
			goto err_out;
		}
		numbytes = ret + strlen(PROCEED_MSG) + 1;
	} else
		numbytes = strlen(buf);
	ret = -E_SEND;
	if (send_bin_buffer(fd, buf, numbytes) < 0)
		goto err_out;

	if (use_rc4) {
		crypt_function_recv = rc4_recv;
		crypt_function_send = rc4_send;
		PARA_INFO_LOG("%s", "rc4 encrytion activated\n");
	}
	/* read command */
	while ((numbytes = recv_buffer(fd, buf, sizeof(buf))) > 0) {
//		PARA_INFO_LOG("recvd: %s (%d)\n", buf, numbytes);
		ret = -E_SYNTAX;
		if (command && numbytes + strlen(command) > STRINGSIZE) /* DOS */
			goto err_out;
		command = para_strcat(command, buf);
		if ((p = strstr(command, EOC_MSG))) {
			*p = '\0';
			break;
		}
	}
	ret = -E_RECV;
	if (numbytes < 0)
		goto err_out;
	ret = -E_CMD_INVALID;
	/* parse command */
	if (!(cmd = parse_cmd(command, &err_list))) {
		err_list = internal_err_list;
		goto err_out;
	}
	/* valid command, check permissions */
	ret = check_perms(u.perms, cmd);
	if (ret < 0) {
		/* insufficient perms */
		err_list = internal_err_list;
		PARA_NOTICE_LOG("insufficient permissions for command %s\n",
			cmd->name);
		goto err_out;
	}
	/* valid command and sufficient perms */
	ret = -E_SPLIT;
	argc = split_args(command, &argv, '\n');
	if (!argv)
		goto err_out;
	argv[0] = cmd->name;
	mmd_lock();
	mmd->num_commands++;
	mmd_unlock();
	PARA_NOTICE_LOG("calling com_%s() for %s@%s\n", cmd->name, u.name,
		inet_ntoa(addr->sin_addr));
	ret = cmd->handler(fd, argc, argv);
	if (ret >= 0) {
		ret = EXIT_SUCCESS;
		goto out;
	}
err_out:
	if (ret != -E_SEND && ret != -E_RECV) {
		PARA_DEBUG_LOG("sending error response #%d (%s)\n",
			-ret, err_list[-ret]);
		send_va_buffer(fd, "%s\n", err_list[-ret]);
	}
	ret = EXIT_FAILURE;
out:
	free(command);
	free(argv);
	mmd_lock();
	if (cmd && (cmd->perms & DB_WRITE) && ret >= 0)
		mmd->events++;
	mmd->active_connections--;
	mmd_unlock();
	return ret;
}
